from importlib.resources import files
from pprint import pprint
import hashlib
from datetime import datetime
from zipfile import ZipFile
import logging
import sys

import click
import xml.etree.ElementTree as ET

from .parse_yaml import parse_yaml
from .board import Board
from .breadboard_svg import breadboard_svg
from .pcb_svg import pcb_svg
from .schematic_svg import schematic_svg
from .fzp_xml import fzp_xml

types: {
    'dict': dict,
    'list': list,
    'str': (str, float, int), # Mainly for 'version', which could be any of these
    'float': (float, int),
    'int': int
}
    
@click.command()
@click.argument('input', type=click.File('rb'))
@click.option('-d', '--defaults', type=click.File('rb'),
              default=files(__package__).joinpath('defaults.yaml'),
              help='Alternative defaults file')
@click.option('-g', '--grid', is_flag=True,
              help='Add gridlines to all SVG images')
@click.option('-b', '--breadboard/--no-breadboard', default=False,
              help='Write breadboard SVG as separate file')
@click.option('-s', '--schematic/--no-schematic', default=False,
              help='Write schematic SVG as separate file')
@click.option('-p', '--pcb/--no-pcb', default=False,
              help='Write PCB SVG as separate file')
@click.option('-f', '--fzp/--no-fzp', default=False,
              help='Write Fritzing XML part description as separate file')
@click.option('-z', '--fzpz/--no-fzpz', default=True, show_default=True,
              help='Write Fritzing FZPZ part as separate file')
@click.option('-a', '--all/--no-all', default=False,
              help='Write everything as files. Same as -b -s -p -f -z')
@click.pass_context
def cli(ctx, input, defaults, grid, all, **write):
    """Convert an INPUT YAML file into a Fritzing part.
    
    Fritzing (https://fritzing.org/) is a tool for designing electronics projects, and includes a view of how a project will look when prototyped using a breadboard. It comes with many parts pre-installed, but you may find it does not include those random small PCBs you bought of some internet site.
    
    Fzpcb is a tool for that takes a YAML input file that describes such a PCB, and converts into a FZPZ part that can be imported into Fritzing.
    
    Fzpcb is not meant for designing PCBs, but rather for producing Fritzing parts that are rough approximations of existing PCBs, to make it easier to check your project prototype is correctly wired before turning it on.
    
    By default it will create the 3 required SVG images and the FZP part description XML, and combine them into a single ZFZP zipfile in the current working dirctory. This can then be imported into Fritzing. CLI options can be used to write one of more of these as standalone files.
    """
    
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    log = logging.getLogger(__name__)
    
    if all:
        for opt in write:
            if ctx.get_parameter_source(opt).name == 'DEFAULT':
                write[opt] = True
            
    log.info(f'Parsing YAML')
    config = parse_yaml(defaults, input)
    log.info(f'Setting missing values to defaults')
    set_missing_values(config, input.name)

    log.info(f'Processing input')
    board = Board(config)
    
    # pprint(config)
    # print(board.layout)

    def et_to_binary(et, encoding):
        return ET.tostring(et, encoding=encoding, xml_declaration=True) + b'\n'

    if write['fzpz']:
        zipname = f"{config['metadata']['basename']}.fzpz"
        log.info(f"Opening zipfile '{zipname}' for writing")
        zipfile = ZipFile(zipname, 'w')

    breadboard_filename = f"svg.breadboard.{config['metadata']['SVGbasename']}_breadboard.svg"
    log.info(f"Constructing breadboard SVG '{breadboard_filename}'")
    xml = breadboard_svg(board, gridlines=grid)
    ET.indent(xml)
    encoded = et_to_binary(xml, config['metadata']['encoding'])
    if write['fzpz']:
        zipfile.writestr(breadboard_filename, encoded)
    if write['breadboard']:
        with open(breadboard_filename, 'wb') as f:
            f.write(encoded)

    pcb_filename = f"svg.pcb.{config['metadata']['SVGbasename']}_pcb.svg"
    log.info(f"Constructing PCB SVG '{pcb_filename}'")
    xml = pcb_svg(board, gridlines=grid)
    ET.indent(xml)
    encoded = et_to_binary(xml, config['metadata']['encoding'])
    if write['fzpz']:
        zipfile.writestr(pcb_filename, encoded)
    if write['pcb']:
        with open(pcb_filename, 'wb') as f:
            f.write(encoded)

    schematic_filename = f"svg.schematic.{config['metadata']['SVGbasename']}_schematic.svg"
    log.info(f"Constructing schematic SVG '{schematic_filename}'")
    xml = schematic_svg(board, gridlines=grid)
    ET.indent(xml)
    encoded = et_to_binary(xml, config['metadata']['encoding'])
    if write['fzpz']:
        zipfile.writestr(schematic_filename, encoded)
    if write['schematic']:
        with open(schematic_filename, 'wb') as f:
            f.write(encoded)

    fzp_filename = f"part.{config['metadata']['SVGbasename']}.fzp"
    log.info(f"Constructing FZP XML '{fzp_filename}'")
    xml = fzp_xml(board, input.name, schematic_filename, pcb_filename, breadboard_filename)
    ET.indent(xml)
    encoded = et_to_binary(xml, config['metadata']['encoding'])
    if write['fzpz']:
        zipfile.writestr(fzp_filename, encoded)
    if write['fzp']:
        with open(fzp_filename, 'wb') as f:
            f.write(encoded)

    if write['fzpz']:
        zipfile.close()
        log.info(f"Wrote zipfile '{zipname}'")

def overlay(defaults, config):
    # Overlay the given config dict on top of the defaults dict
    # FIXME: Could do way more error checking of config
    for key, value in config:
        if key in defaults:
            if isinstance(defaults[key], dict):
                if not isinstance(value, dict):
                    raise ValueError(f'Expected {key} to be a dictionary')
                else:
                    overlay(defaults[key], value)
            elif isinstance(defaults[key], list):
                if not isinstance(value, list):
                    raise ValueError(f'Expected {key} to be a list')
                else:
                    defaults[key] = value
            elif isinstance(defaults[key], str):
                if not isinstance(value, (str, float, int)):
                    raise ValueError(f'Expected {key} to be a string')
                else:
                    defaults[key] = str(value)
            elif isinstance(defaults[key], float):
                if not isinstance(value, (float, int)):
                    raise ValueError(f'Expected {key} to be a float')
                else:
                    defaults[key] = float(value)
            elif isinstance(defaults[key], int):
                if not isinstance(value, int):
                    raise ValueError(f'Expected {key} to be an integer')
                else:
                    defaults[key] = str(value)
            elif isinstance(defaults[key], bool):
                if not isinstance(value, bool):
                    raise ValueError(f'Expected {key} to be an boolean')
                else:
                    defaults[key] = value
            else:
                raise ValueError(f'Unexpected type {type(defaults[key])} for default {key}, value: "{defaults[key]}"')
    else:
        defaults[key] = value

def set_missing_values(config, filename):
    # Fill in optional values missing from the input file, and that have
    # no defaullt
    log = logging.getLogger(__name__)

    if 'moduleId' not in config['metadata']:
        md5str = f'{config["metadata"]["author"]}|{config["metadata"]["version"]}|{config["metadata"]["title"]}|{filename}'
        
        hexdigest = hashlib.md5(md5str.encode('utf-8')).hexdigest()
        config['metadata']['moduleId'] = hexdigest
        log.info(f"Set moduleId to '{config['metadata']['moduleId']}'")
        
    if 'description' not in config['metadata']:
        config['metadata']['description'] = config['metadata']['title']
        log.info(f"Set description to '{config['metadata']['description']}'")
        
    if 'label' not in config['metadata']:
        config['metadata']['label'] = config['metadata']['title']
        log.info(f"Set label to '{config['metadata']['label']}'")
        
    if 'basename' not in config['metadata']:
        config['metadata']['basename'] = ''.join(filter(str.isalnum, config['metadata']['label']))
        log.info(f"Set basename to '{config['metadata']['basename']}'")

    if 'SVGbasename' not in config['metadata']:
        config['metadata']['SVGbasename'] = config['metadata']['basename'] + '_' + config['metadata']['moduleId']
        log.info(f"Set SVGbasename to '{config['metadata']['SVGbasename']}'")

    try:
        config['metadata']['date'] = config['metadata']['date'].strftime('%Y-%m-%d')
    except AttributeError:
        config['metadata']['date'] = str(config['metadata']['date'])
    except KeyError:
        config['metadata']['date'] = datetime.today().strftime('%Y-%m-%d')
    log.info(f"Set date to '{config['metadata']['date']}'")



#!/usr/bin/env python

from enum import Enum, auto #, IntFlag
from dataclasses import dataclass, astuple
# from math import atan2, cos, sin, pow, sqrt
from copy import copy

class Dir(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()
#     
# class Edge(IntFlag):
#     NONE = 0
#     TOP = 1
#     BOTTOM = 2
#     LEFT = 4
#     # TOP_LEFT = 5
#     # BOTTOM_LEFT = 6
#     RIGHT = 8
#     # TOP_RIGHT = 9
#     # BOTTOM_RIGHT = 10
#     ALL = 15
#     
#     def flip(self):
#         # Returns new edge with left and right swapped
#         
#         # Edge will be immutable, so no need to copy
#         # This can probably be done faster, but who cares?
#         left = self & self.__class__.LEFT
#         self ^= left
#         right = self & self.__class__.RIGHT
#         self ^= right
#         if left:
#             self |= self.__class__.RIGHT
#         if right:
#             self |= self.__class__.LEFT
#         return self
#     
#     def flop(self):
#         # Returns new edge with top and bottom swapped
#         top = self & self.__class__.TOP
#         self ^= top
#         bottom = self & self.__class__.BOTTOM
#         self ^= bottom
#         if top:
#             self |= self.__class__.BOTTOM
#         if bottom:
#             self |= self.__class__.TOP
#         return self
#     
#     def chain(self, ops):
#         new = copy(self)
#         for op in ops:
#             new = op(new)
#         return new

class Coords:
    def __getitem__(self, item):
        return astuple(self)[item]
    
    # def __str__(self):
    #     # String reprsentation is just the two coords separated by a space,
    #     # as used by SVG in paths, viewboxes, etc
    #     # FIXME Not needed?
    #     return f'{self[0]} {self[1]}'
    
    def __mul__(self, multiplier):
        # print(f'{self}.__mul__({multiplier})')
        try:
            return self.__class__(self[0] * multiplier[0], self[1] * multiplier[1])
        except TypeError:
            return self.__class__(self[0] * multiplier, self[1] * multiplier)
        
    __rmul__ = __mul__
    
    def __truediv__(self, divisor):
        try:
            return self.__class__(self[0] / divisor[0], self[1] / divisor[1])
        except TypeError:
            return self.__class__(self[0] / divisor, self[1] / divisor)
        
    def __add__(self, other):
        # print(repr(self))
        # print(repr(other))
        try:
            return self.__class__(self[0] + other[0], self[1] + other[1])
        except TypeError:
            return self.__class__(self[0] + other, self[1] + other)
    
    def __sub__(self, other):
        try:
            return self.__class__(self[0] - other[0], self[1] - other[1])
        except TypeError:
            return self.__class__(self[0] - other, self[1] - other)

    def __neg__(self):
        return self.__class__(-self[0], -self[1])

    def __abs__(self):
        return self.__class__(abs(self[0]), abs(self[1]))
    
    def maxwith(self, other):
        try:
            return self.__class__(max(self[0], other[0]), max(self[1], other[1]))
        except TypeError:
            return self.__class__(max(self[0], other), max(self[1], other))
    
    def minwith(self, other):
        try:
            return self.__class__(min(self[0], other[0]), min(self[1], other[1]))
        except TypeError:
            return self.__class__(min(self[0], other), min(self[1], other))

    def across(self):
        # Only the x / width / dx coordinate
        return self.__class__(self[0], 0)
    
    def right(self):
        # Only the x / width / dx coordinate
        return self.__class__(self[0], 0)
    
    def left(self):
        # Only the y / height / dy coordinate, negated
        return -self.right()
    
    def down(self):
        # Only the y / height / dy coordinate
        return self.__class__(0, self[1])

    def up(self):
        # Only the y / height / dy coordinate, negated
        return -self.down()

    def flip(self):
        # Invert the x / width / dx coordinate
        return self.__class__(-self[0], self[1])

    def flop(self):
        # Invert the y / height / dy coordinate
        return self.__class__(self[0], -self[1])
    
    def flipflop(self):
        # Invert both coords. Same as __neg__
        return self.__class__(-self[0], -self[1])
    
    def identity(self):
        return self
    
    # flip and flop together can be done with -coord (or -1 * coord)
    
    def transpose(self):
        # Swap coordinates, transpose+flip and transpose+flop are rotations
        return self.__class__(self[1], self[0])

    def length(self):
        return sqrt(pow(self[0], 2) + pow(self[1], 2))
    
    def angle(self):
        return atan2(self[1], self[0])
    
    def point_on_ellipse(self, angle):
        # Give point 'angle' way around an ellipse with semi-axies of coords
        ab = self[0] * self[1]
        b_cos = self[1] * cos(angle)
        a_sin = self[0] * sin(angle)
        div = sqrt(pow(b_cos, 2) + pow(a_sin, 2))
        x = ab * cos(angle) / div
        y = ab * sin(angle) / div
        return self.__class__(x, y)
    
    def chain(self, ops):
        new = copy(self)
        for op in ops:
            new = new.__getattribute__(op)()
        return new

    def pchain(self, ops):
        new = copy(self)
        for op in ops:
            if op == 'transpose':
                new = new.__getattribute__(op)()
        return new

    def fchain(self, ops):
        new = self
        for op in ops:
            large_arc_flag = new[0]
            sweep_flag = abs(new[1] - 1)
            new = Vector(large_arc_flag, sweep_flag)
        return new


    # FIXME do we need/want rotations?
    
    @classmethod
    def zero(cls):
        return cls(0, 0)

    @classmethod
    def half(cls):
        return cls(0.5, 0.5)

    @classmethod
    def one(cls):
        return cls(1, 1)

    @classmethod
    def two(cls):
        return cls(2, 2)

@dataclass
class Position(Coords):
    x: float
    y: float

@dataclass
class Size(Coords):
    width: float
    height: float

@dataclass
class Vector(Coords):
    dx: float
    dy: float
    
# @dataclass
# class ArcFlags(Coords):
#     # SVG arc flags that understand what to do with flip and flop
#     large_arc_flag: int
#     sweep_flag: int
#             
#     def flip(self):
#         return self.__class__(self.large_arc_flag, int(not self.sweep_flag))
#     
#     mirror = flop = flip
#     
#     # def mirror(self):
#     #     return self
    
class CoordsTo:
    def __init__(self, border, cell_position, scale, part_size=None, ops=None, step=1):
        self.border = border
        self.cell_position = cell_position
        self.scale = scale
        self.step = step
        if part_size:
            self.part_size = part_size
        else:
            self.part_size = Size(1,1)
        if ops:
            self.ops = ops
        else:
            self.ops = []
        
    def offset(self, offset):
        new = copy(self)
        new.border += offset
        return new
    
    def __repr__(self):
        return f'{self.__class__.__name__}(border={self.border}, cell_position={self.cell_position}, scale={self.scale}, part_size={self.part_size})'
    
    def cabs(self, coords):
        cell_centre = (self.border + self.cell_position + Vector.half()) * self.scale
        offset_from_centre = coords - Vector.half()
        result = cell_centre + self.crel(offset_from_centre)
        # print('cabs', coords, '->', result / self.scale)
        return result
    
    def crel(self, coords):
        result = coords.chain(self.ops) * self.scale
        # print('crel', coords, '->', result / self.scale)
        return result
    
    def clen(self, length):
        return length * self.scale
    
    def sabs(self, coords):
        step_centre = (self.border + self.cell_position + Vector.half() * self.step) * self.scale
        offset_from_centre = (coords - Vector.half()) * self.step
        result = step_centre + self.crel(offset_from_centre)
        # print('sabs', coords, '->', result / self.scale)
        return result

    # def sabs(self, coords):
    #     result = self.cabs(coords * self.step)
    #     # print('cabs', coords, '->', result / self.scale)
    #     return result
    
    def srel(self, coords):
        result = self.crel(coords * self.step)
        # print('crel', coords, '->', result / self.scale)
        return result
    
    def slen(self, length):
        return self.clen(length * self.step)

    def pabs(self, coords):
        # print('coords', coords)
        part_size = self.part_size.pchain(self.ops)
        offset = Vector.one()
        # print('half', half)
        part_centre = (self.border + self.cell_position + offset) * self.scale
        # print('part_centre', part_centre / self.scale)
        offset_from_centre = coords * part_size - offset
        # print('offset_from_centre', offset_from_centre)
        # result = part_centre + self.prel(offset_from_centre)
        result = part_centre + offset_from_centre.chain(self.ops) * self.scale
        # print('pabs', coords, '->', coords * self.part_size, '->', result / self.scale)
        # print(self)
        return result
    
    def prel(self, coords):
        result = self.crel(coords * self.part_size.pchain(self.ops))
        # print('prel', coords, '->', coords * self.part_size, '->', result / self.scale)
        return result
    
    def prelh(self, coords):
        transposed_size = self.part_size.pchain(self.ops)
        part_size_h = Size(transposed_size[0], transposed_size[0])
        result = self.crel(coords * part_size_h.pchain(self.ops))
        # print('prel', coords, '->', coords * self.part_size, '->', result / self.scale)
        return result

    def prelv(self, coords):
        transposed_size = self.part_size.pchain(self.ops)
        part_size_v = Size(transposed_size[1], transposed_size[1])
        result = self.crel(coords * part_size_v.pchain(self.ops))
        # print('prel', coords, '->', coords * self.part_size, '->', result / self.scale)
        return result

    def parcsize(self, coords):
        result = self.crel(coords * self.part_size.pchain(self.ops))
        result = coords.__class__(abs(result[0]), abs(result[1]))
        # print('prel', coords, '->', coords * self.part_size, '->', result / self.scale)
        return result
    
    def transpose_only(self, coords):
        # print(self.ops)
        result = coords.chain(self.ops)
        result = coords.__class__(abs(result[0]), abs(result[1]))
        # print('prel', coords, '->', coords * self.part_size, '->', result / self.scale)
        return result

    def parcflags(self, flags):
        return flags.fchain(self.ops)
    
    def plenh(self, length):
        return self.clen(length * self.part_size.pchain(self.ops)[0])
    
    def plenv(self, length):
        return self.clen(length * self.part_size.pchain(self.ops)[1])
    
    def plen(self, length):
        return self.clen(length * (self.part_size[0] + self.part_size[1])/2)
    
    def jump(self, coords):
        return self.crel(coords + (self.part_size - Vector.one()).pchain(self.ops))
    
    def jumpback(self, coords):
        return self.crel(coords - (self.part_size + Vector.one()).pchain(self.ops))
    
    def jumpright(self, coords):
        # print('A', coords)
        # print('B', self.part_size)
        # print('C', Vector(self.part_size[0]-1, 0))
        # print('D', coords + Vector(self.part_size[0]-1, 0).pchain(self.ops))
        return self.crel(coords + Vector(self.part_size.pchain(self.ops)[0]-1, 0))
    
    def jumpdown(self, coords):
        return self.crel(coords + Vector(0, self.part_size.pchain(self.ops)[1]-1))
    
    def jumpleft(self, coords):
        return self.crel(coords - Vector(self.part_size.pchain(self.ops)[0]-1, 0))
    
    def jumpup(self, coords):
        return self.crel(coords - Vector(0, self.part_size.pchain(self.ops)[1]-1))
    
    def f(self, arc_flags):
        # FIXME What happens if self.ops are subclassed from Coords.flip, etc.
        # but coords is an ArcFlag
        return arc_flags.chain(self.ops)
    
    # def chain(self, ops):
    #     # Apply ops to coord
    #     new = copy(self)
    #     for op in ops:
    #         new = op(new)
    #     return new
#         
#     def pchain(self, ops):
#         # Only apply transpose to coord, necessary to recalc partsize
#         new = copy(self)
#         for op in ops:
#             if op == 'transpose':
#                 new = op(new)
#         return new
        
    def expand_args_list(self, args_list):
        # Convert list of part functions arguments to real-world coordinates
        
        one_arg_funcs = ['clen', 'slen', 'plenh', 'plenv', 'plen']
        two_arg_funcs = ['pabs', 'prel', 'prelh', 'prelv', 'parcsize', 'parcflags', 'cabs', 'crel', 'sabs', 'srel', 'jump', 'jumpback', 'jumpright', 'jumpdown', 'jumpleft', 'jumpup']
        result = []
        seven_arg_funcs = ['parc'] #, 'pcirc']
        
        for arg in args_list:
            if isinstance(arg, dict):
                if len(arg) != 1:
                    raise ValueError(f'Part function maps must have only one key: {arg}')
                
                for func, func_args in arg.items(): # There is only one, Highlander
                    if func in one_arg_funcs:
                        if len(func_args) != 1:
                            raise ValueError(f'Part function {func} expects 1 argument: {arg}')
                    elif func in two_arg_funcs:
                        if len(func_args) != 2:
                            raise ValueError(f'Part function {func} expects 2 arguments: {arg}')
                    elif func in seven_arg_funcs:
                        if len(func_args) != 7:
                            raise ValueError(f'Part function {func} expects 7 arguments: {arg}')
                    else:
                        raise ValueError(f'Unrecognzed part function: {func}')

                    if func == 'clen':
                        result.append(self.clen(*func_args))
                    elif func == 'slen':
                        result.append(self.slen(*func_args))
                    elif func == 'plenh':
                        result.append(self.plenh(*func_args))
                    elif func == 'plenv':
                        result.append(self.plenv(*func_args))
                    elif func == 'plen':
                        result.append(self.plen(*func_args))
                    elif func == 'pabs':
                        result.extend(self.pabs(Vector(*func_args)))
                    elif func == 'prel':
                        result.extend(self.prel(Vector(*func_args)))
                    elif func == 'prelh':
                        result.extend(self.prelh(Vector(*func_args)))
                    elif func == 'prelv':
                        result.extend(self.prelv(Vector(*func_args)))
                    elif func == 'parcsize':
                        result.extend(self.parcsize(Vector(*func_args)))
                    elif func == 'parcflags':
                        result.extend(self.parcflags(Vector(*func_args)))
                    elif func == 'parc':
                        result.extend([
                            *self.parcsize(Vector(func_args[0], func_args[1])),
                            func_args[2],
                            *self.parcflags(Vector(func_args[3], func_args[4])),
                            *self.prel(Vector(func_args[5], func_args[6]))])
                    # elif func == 'pcirc':
                    #     avg = func_args[0] + func_args[1] / 2
                    #     result.extend([
                    #         *self.parcsize(Vector(avg, avg)),
                    #         func_args[2],
                    #         *self.parcflags(Vector(func_args[3], func_args[4])),
                    #         *self.prel(Vector(func_args[5], func_args[6]))])
                    elif func == 'cabs':
                        result.extend(self.cabs(Vector(*func_args)))
                    elif func == 'crel':
                        result.extend(self.crel(Vector(*func_args)))
                    elif func == 'sabs':
                        result.extend(self.sabs(Vector(*func_args)))
                    elif func == 'srel':
                        result.extend(self.srel(Vector(*func_args)))
                    elif func == 'jump':
                        result.extend(self.jump(Vector(*func_args)))
                    elif func == 'jumpleft':
                        result.extend(self.jumpleft(Vector(*func_args)))
                    elif func == 'jumpright':
                        result.extend(self.jumpright(Vector(*func_args)))
                    elif func == 'jumpup':
                        result.extend(self.jumpup(Vector(*func_args)))
                    elif func == 'jumpdown':
                        result.extend(self.jumpdown(Vector(*func_args)))
                    else:
                        raise AssertionError("Can't happen")
                            
            elif isinstance(arg, (str, float, int)):
                result.append(arg)
            else:
                raise ValueError(f'Part args can only be dict, str, float, or int: "{arg}"')
        return result

if __name__ == '__main__':
    position = Position(0.2, 0.3)
    print(f'position = {position}')
    print(f'position with str = {str(position)}')
    print(f'position with repr = {repr(position)}')
    vector = Vector(3, 4)
    print(f'vector = {vector}')
    scale = Size(10, 20)
    print(f'scale = {scale}')
    
    c2 = CoordsTo(Size(1, 1), Position(2, 3), Size(10, 10))
    print(f'c2.a(position) = {c2.a(position)}')
    print(f'c2.r(position) = {c2.r(position)}')

    edges = Edge.LEFT | Edge.TOP
    print(edges)
    print(edges.flip())
    print(edges.flip().flop())
    
    print()
    print('Testing chain')
    print(edges)
    ops = [Edge.flip, Edge.flop]
    print(edges.chain(ops))
    
    sf = SweepFlag(2)
    print(sf)
    print(sf.flip())
    print(sf.flip().flop())
    



from pprint import pprint

import yaml

import logging
log = logging.getLogger(__name__)

def _overlay(defaults, config):
    # FIXME: Could do way more error checking of config
    for key, value in config.items():
        if key in defaults:
            if isinstance(defaults[key], dict):
                if not isinstance(value, dict):
                    raise ValueError(f'Expected {key} to be a dictionary')
                else:
                    _overlay(defaults[key], value)
            elif isinstance(defaults[key], list):
                if not isinstance(value, list):
                    raise ValueError(f'Expected {key} to be a list')
                else:
                    defaults[key] = value
            elif isinstance(defaults[key], str):
                if not isinstance(value, (str, float, int)):
                    raise ValueError(f'Expected {key} to be a string')
                else:
                    defaults[key] = str(value)
            elif isinstance(defaults[key], float):
                if not isinstance(value, (float, int)):
                    raise ValueError(f'Expected {key} to be a float')
                else:
                    defaults[key] = float(value)
            elif isinstance(defaults[key], int):
                if not isinstance(value, int):
                    raise ValueError(f'Expected {key} to be an integer')
                else:
                    defaults[key] = str(value)
            elif isinstance(defaults[key], bool):
                if not isinstance(value, bool):
                    raise ValueError(f'Expected {key} to be an boolean')
                else:
                    defaults[key] = value
            else:
                raise ValueError(f'Unexpected type {type(defaults[key])} for default {key}, value: "{defaults[key]}"')
        else:
            defaults[key] = value

def parse_yaml(defaults_fp, config_fp):
    log.info(f"Reading defaults from '{defaults_fp.name}'")
    defaults = yaml.load(defaults_fp, yaml.Loader)
    log.info(f"Reading input from '{config_fp.name}'")
    config = yaml.load(config_fp, yaml.Loader)
#     
#     print('defaults')
#     print('========')
#     pprint(defaults)
#     print()
# 
#     print('config')
#     print('======')
#     pprint(config)
#     print()
    
    _overlay(defaults, config)
    
    return defaults
        
        



from copy import copy

def find_colour_from_key(colour_dict, colour_key, default_key='unknown'):
    # Given a colour dict and a key, find the color
    # colour_key may be a tuple/list of key and index, in which case
    # return that specific colour, not the first
    if isinstance(colour_key, (tuple, list)):
        index = colour_key[1]
        colour_key = colour_key[0]
    else:
        index = 0

    if colour_key == '?':
        colour_key = default_key
        
    try:
        colour = colour_dict[colour_key]
    except KeyError:
        colour = colour_dict['unknown']

    return colour if index == 0 else colour[index]

def find_part_code_from_key(parts_dict, part_key):
    # Given a part code dict and a key, find the part code
    # Can be copies or transposed copies of other parts
    part_ops = []
    try:
        part_code = parts_dict[part_key]
        try:
            transpose_of = part_code['transpose']
            try:
                name = parts_dict[part_key]['name']
                part_code = parts_dict[transpose_of]
                part_code['name'] = name
            except KeyError:
                raise ValueError(f'Part {part_key} tried to transpose non-existant part {transpose_of}')
            part_ops.append('transpose')
        except KeyError:
            pass
            
        try:
            # Copy beats transpose, for no particular reason
            copy_of = part_code['copy']
            try:
                name = parts_dict[part_key]['name']
                part_code = parts_dict[copy_of]
                part_code['name'] = name
            except KeyError:
                raise ValueError(f'Part {part_key} tried to copy non-existant part {copy_of}')
        except KeyError:
            pass

    except KeyError:
        part_code = None # Special drawing code for this case
        
    return part_code, part_ops

def merge_subdicts_with_defaults(parent, starts_with, key_for_strings='text'):
    # Merge subdicts in parent that have a key starting with starts_with
    merge_dict_names = [k for k in parent.keys() if k.startswith(starts_with)]
    result = {}
    for merge_dict_name in merge_dict_names:
        merge_dict = parent[merge_dict_name]
        try:
            defaults = merge_dict['defaults']
        except KeyError:
            print(merge_dict_name)
            print(merge_dict)
            raise ValueError(f'Dictionary {merge_dict_name} needs a "defaults" key')
        for key, value in merge_dict.items():
            if key == 'defaults':
                continue
            if not isinstance(value, dict):
                text = str(value)
                value = copy(defaults)
                value[key_for_strings] = text
            for def_key, def_value in defaults.items():
                if def_key not in value:
                    value[def_key] = def_value
            result[key] = value
        del(parent[merge_dict_name])
    return result



from copy import copy
from collections import OrderedDict

from .colour_list import ColourList
from .layout import Layout
from .coords import Position, Size
from .helpers import find_colour_from_key, merge_subdicts_with_defaults
from .array_2d import Array2D
from .element import Element
from .cell import ConnectorCell, PartCell, TextCell

class Board():
    # Convert into the format we work with
    def __init__(self, config):
        self.config = config

        # Turn colour strings into ColoursLists
        # print(self.config['colours'])
        new_colour_dict = {}
        for colour_code, sequence in self.config['colours'].items():
            # print(f'{colour_code} mapped to {sequence}')
            colour_list = ColourList(sequence)
            try:
                colour_letter, colour_word = colour_code.split('+', maxsplit=1)
            except ValueError:
                new_colour_dict[colour_code] = colour_list
            else:
                new_colour_dict[colour_letter] = colour_list
                new_colour_dict[colour_word] = colour_list
        self.config['colours'] = new_colour_dict
        # for k, v in self.config['colours'].items():
        #     try:
        #         print(k, repr(v))
        #     except TypeError:
        #         print(f"Can't print {k}")
        # print(self.config['colours']['pcb'])
        
        # Extract colour_codes that indicate text
        text_chars = []
        for letter, colour_list in config['colours'].items():
            if colour_list.istext():
                text_chars.append(letter)
                
        self.source = Array2D([Element.from_str(s) for s in config['layout']])
        # Unlink empty cells, so outline of PCB can be traced
        for element in self.source.flatten():
            if element.char == config['specials']['empty']:
                element.unlink()

        self.layout = Layout(self.source, 
                             config['specials']['empty'],
                             config['specials']['pcb'],
                             text_chars,
                             config['extenders']['top'],
                             config['extenders']['middle'],
                             config['extenders']['half-bottom'])

        # Find the connector cells, and put them in a ordered dict
        connectors = {}
        for cell in self.layout.flatten():
            if isinstance(cell, ConnectorCell):
                if cell.number in connectors:
                    raise ValueError(f'Duplicate connector {cell.number} at position {cell.element.position.x},{cell.element.position.y}')
                else:
                    connectors[cell.number] = cell
        self.connectors = OrderedDict()
        for cell in [connectors[key] for key in sorted(connectors.keys())]:
            self.connectors[cell.number] = cell

        # for number, (key, connector) in enumerate(self.connectors.items()):
        #     print(number, key, connector)
        
        # Find the part cells, and put them in a list
        self.parts = []
        for cell in self.layout.flatten():
            if isinstance(cell, PartCell):
                self.parts.append(cell)

        # Find the text cells, and put them in a dict
        self.text = {}
        for cell in self.layout.flatten():
            if isinstance(cell, TextCell):
                if cell.text_key in self.text:
                    raise ValueError(f'Duplicate text {cell.text_key} at position {cell.element.position.x},{cell.element.position.y}')
                else:
                    self.text[cell.text_key] = cell

        # Apply defaults to connector dicts, and merge them
        connector_defaults = config['connectors']['defaults']
        connector_dict = merge_subdicts_with_defaults(config, 'connector')
         
        # Take the connector dict values, and use them to update the
        # connector cell
        for cid, cell in enumerate(self.connectors.values()):
            try:
                connector = connector_dict[cell.number]
            except KeyError:
                connector = connector_defaults
            self.update_connector(cid, cell, connector)
            
        # Find connectors with the same name and put them in buses
        conn_by_label = {}
        for cell in self.connectors.values():
            try:
                conn_by_label[cell.label].append(cell)
            except KeyError:
                conn_by_label[cell.label] = [cell]
        self.buses = {}
        for label, cell_list in conn_by_label.items():
            if len(cell_list) > 1:
                self.buses[label] = cell_list
        
        # Apply defaults to text dicts, and merge them
        self.text_dict = merge_subdicts_with_defaults(config, 'text')
                
        # Add undefined text keys to the text dict
        for cell in self.text.values():
            if cell.text_key not in self.text_dict:
                self.text_dict[cell.text_key] = config['text']['defaults']
                self.text_dict[cell.text_key]['text'] = f'Text-{cell.text_key}-undefined'

        self.size = self.layout.size
        # print(self.text_dict)

    @staticmethod
    def update_connector(cid, cell, connector):
        # Update the a cell with connector details, now we know them
        colour_key = connector['colour']
        gender = connector['gender']

        try:
            text = connector['text']
        except KeyError:
            text = f'PIN{cell.number} Connector {cell.number}'

        pair = text.split(' ', maxsplit=1)
        
        try:
            label = connector['label']
        except KeyError:
            try:
                label = pair[0]
            except IndexError:
                label = ''

        try:
            description = connector['description']
        except KeyError:
            try:
                description = pair[1]
            except IndexError:
                description = ''

        # Work out if it is an edge connector, and of so, what the operations
        # to perform to rotate the part correctly
        # NOTE Only elements are unlinked if empty, not cells
        part_ops = []
        edge = False
        if not cell.element.left:
            part_ops = []
            edge = True
        elif not cell.element.right or not cell.element.right.right:
            part_ops = ['flip']
            edge = True
        elif not cell.element.up:
            part_ops = ['transpose']
            edge = True
        elif not cell.element.down:
            part_ops = ['transpose', 'flop']
            edge = True

        # Work out what possible parts to use
        if label == 'GND':
            if edge:
                possible = ['gnd-edge-part', 'gnd-part', 'edge-part', 'part']
            else:
                possible = ['gnd-part', 'part']
        else:
            if edge:
                possible = ['edge-part', 'part']
            else:
                possible = ['part']

        # And pick the first one that matches:
        for part_key_name in possible:
            if part_key_name in connector:
                part_key = connector[part_key_name]
                break
        else:
            raise ValueError('Connectors must have a (possible default) part')
                            
        cell.update(colour_key, part_key, label, 
                                description, gender, cid, part_ops)


    def colour(self, colour_key, default_key='unknown'):
        # Helper function to find colour
        return find_colour_from_key(self.config['colours'], colour_key, default_key)



from textwrap import wrap

from .coords import Size, Position, Vector
from .grid_drawing import GridDrawing
from .cell import ConnectorCell
# from .helpers import find_colour_from_key

def schematic_svg(board, gridlines=False):
    # Draw PCN SVG, returning it in ElementTree XML
    
    # Start with grid with border of one added (for the pins)
    # - Put any pins on left or right on their sides
    # - If gaps on left, go through half the row (L->R) looking for pin
    # - If gaps in the right, go though all the row (R->L) looking for pin
    # - If gaps on left, go through all the row (L->R) looking for pin
    # - Go through board (scanning down first), putting pins
    #   on the top (not corners). If coallescing GND pins, skip them
    # - If not wide enough, see how many left over, widen by that much
    # - If coallescing, Remove GND pins, and put on bottom middle
    # - Try to space out top pins
    
    left = [None] * board.source.size.height
    right = [None] * board.source.size.height
    top = [None] * board.source.size.width
    
    # Coordinates are like this:
    #         3 |   4 |
    #           |     |
    #      +----x-----x--
    #      |  (4,2) (6,2) [space out if not fully populated, extend else]
    #   1  | 
    # -----X (2,4)       (W,4) X-----
    #      |                   |
    #   2  |                   |
    # -----x (2,6)       (W,6) X----
    #     
    
    right_side = board.source.size.width+2
    
    for top_to_left, top_to_right in zip(board.layout[0], reversed(board.layout[0])):
        for cell_to_left in top_to_left.iter_down():
            if isinstance(cell_to_left, ConnectorCell):
                # print(cell_to_left)
                y = cell_to_left.element.position.y
                if not cell_to_left.schemed and not left[y]:
                    left[y] = cell_to_left
                    cell_to_left.schemed = Position(2, y*2+4)
        for cell_to_right in top_to_right.iter_down():
            if isinstance(cell_to_right, ConnectorCell):
                y = cell_to_right.element.position.y
                if not cell_to_right.schemed and not right[y]:
                    right[y] = cell_to_right
                    cell_to_right.schemed = Position(right_side, y*2+4)
         
    remaining = [cell for cell in board.connectors.values() if not cell.schemed]
    if len(remaining) > board.layout.size.width:
        raise NotImplementedError(f'Need code if too many pins, these are left {remaining}')
    else:
        for nx, cell_to_top in enumerate(remaining):
            x = 2 + (board.layout.size.width) * (nx+1) / (len(remaining)+1)
            top[nx] = cell_to_top
            cell_to_top.schemed = Position(x, 2)
            
    # print([(f'{cell.schemed.x}x{cell.schemed.y}' if cell else 'None') for cell in left])
    # print([(f'{cell.schemed.x}x{cell.schemed.y}' if cell else 'None') for cell in right])
    # print([(f'{cell.schemed.x}x{cell.schemed.y}' if cell else 'None') for cell in top])
    # print(right)
    # print(top)
      
    ############# Draw it all ################
    
    grid = GridDrawing(
        size=board.size + Size(4, 6),
        border=Size(0, 0),
        grid_size=board.config['metadata']['grid_size'],
        grid_unit=board.config['metadata']['grid_unit'],
        scale=board.config['metadata']['scale'],
        parts_dict=None)

    schematic = grid.g(id='schematic')

    c2 = grid.coords_to
    appearance = board.config['appearance']['schematic']
   
    line_col = board.colour(appearance['tick-colour'])
    line_len = appearance['tick']
    line_wid = appearance['tick-width']
    
    schematic.add(grid.rect(id="outline",
                       fill='none',
                       stroke=board.colour(appearance['outline-colour']),
                       stroke_width=c2.clen(appearance['outline-width']),
                       insert=c2.cabs(Position(2, 2)),
                       size=c2.cabs(board.size + Size(0, 2))))

    wrap_at = int((board.source.size.width - 2) / 1.1*appearance['label-font-size'])
    # print(wrap_at)
    lines = wrap(board.config['metadata']['label'], width=wrap_at)
    
    group = grid.g(id=f'label')
    for y, line in enumerate(lines):
        
        insert = Position(board.source.size.width/2+2,
                          board.source.size.height/3+3 +
                          y * 1.5*appearance['name-font-size'] )
        group.add(grid.text(insert=c2.cabs(insert),
                           text_anchor='middle',
                           fill=board.colour(appearance['name-colour']),
                           font_size=c2.clen(appearance['name-font-size']),
                           font_family=appearance['name-font'],
                           text=line))
    schematic.add(group)


    for cell in [cell for cell in left if cell]:
        tick_vector = Vector(-appearance['tick'], 0)
        id_offset = Vector(-appearance['id-font-size']*0.75,
                           -appearance['id-font-size']*0.5)
        label_offset = Vector(appearance['label-font-size']*0.4,
                              appearance['label-font-size']*0.3)
       
        connector = connector_draw(board, grid, cell, appearance, c2, tick_vector,
                                   id_offset, 'end', label_offset, 'start')
        schematic.add(connector)
        
    for cell in [cell for cell in right if cell]:
        tick_vector = Vector(appearance['tick'], 0)
        id_offset = Vector(appearance['id-font-size']*0.75,
                           -appearance['id-font-size']*0.5)
        label_offset = Vector(-appearance['label-font-size']*0.4,
                              appearance['label-font-size']*0.3)
       
        connector = connector_draw(board, grid, cell, appearance, c2, tick_vector,
                                   id_offset, 'start', label_offset, 'end')
        schematic.add(connector)

    for cell in [cell for cell in top if cell]:
        tick_vector = Vector(0, -appearance['tick'])
        id_offset = Vector(-appearance['id-font-size']*0.5,
                           -appearance['id-font-size']*0.75)
        label_offset = Vector(0,
                              appearance['label-font-size']*1.1)
       
        connector = connector_draw(board, grid, cell, appearance, c2, tick_vector,
                                   id_offset, 'end', label_offset, 'middle')
        schematic.add(connector)

    grid.add(schematic)

    if gridlines:
        grid.add(grid.grid_lines(*board.size))

    return grid.get_xml()

def connector_draw(board, grid, cell, appearance, c2, tick_vector, id_offset, 
                   id_text_anchor, label_offset, label_text_anchor):
    
    group = grid.g(id=f'connector{cell.cid}pin')
        
    group.add(grid.line(stroke=board.colour(appearance['tick-colour']),
                    stroke_width=c2.clen(appearance['tick-width']),
                    start=c2.cabs(cell.schemed),
                    end=c2.cabs(cell.schemed + tick_vector)))
    
    group.add(grid.text(insert=c2.cabs(cell.schemed + id_offset),
                    text_anchor=id_text_anchor,
                    fill=board.colour(appearance['id-colour']),
                    font_size=c2.clen(appearance['id-font-size']),
                    font_family=appearance['id-font'],
                    text=str(cell.number)))

    group.add(grid.text(insert=c2.cabs(cell.schemed + label_offset),
                    text_anchor=label_text_anchor,
                    fill=board.colour(appearance['label-colour']),
                    font_size=c2.clen(appearance['label-font-size']),
                    font_family=appearance['label-font'],
                    text=cell.label))

    terminal = Size(appearance['terminal-size'], appearance['terminal-size'])
    group.add(grid.rect(id=f'connector{cell.cid}terminal',
                        fill='none',
                        insert=c2.cabs(cell.schemed + tick_vector - terminal/2),
                        size=c2.crel(terminal)))

    return group



#!/usr/bin/env python
import itertools

from .coords import Position, Size

class Member():
    # Member of a 2D array that knows its position and neighbours
    # Expected to be sub-classed
    
    def __init__(self):
        self.position = None
        self.array_1d = None
        self.up = None
        self.down = None
        self.left = None
        self.right = None
    
    def __str__(self):
        up = (self.up.atstr() if self.up else 'nul')
        down = (self.down.atstr() if self.down else 'nul')
        left = (self.left.atstr() if self.left else 'nul')
        right = (self.right.atstr() if self.right else 'nul')
        return f'@{self.atstr()} <{left} ^{up} v{down} >{right}'
    
    def atstr(self):
        return f'{self.position.x},{self.position.y}'
        
    def unlink(self):
        if self.up:
            # print(f'Unlink from {self.up.atstr()}')
            self.up.down = None
            self.up = None
        if self.down:
            # print(f'Unlink from {self.down.atstr()}')
            self.down.up = None
            self.down = None
        if self.left:
            # print(f'Unlink from {self.left.atstr()}')
            self.left.right = None
            self.left = None
        if self.right:
            # print(f'Unlink from {self.right.atstr()}')
            self.right.left = None
            self.right = None
    
    # FIXME Would be nice to have swap(self, other), but links get tricky
    # when swapping neighbours
    def swap_below(self):
        if not self.down:
            raise AssertionError(f'Can swap member @{self.atstr()} with one below')
        
        other = self.down
        
        # Swap attributes
        for attr in ['position', 'array_1d', 'up', 'down', 'left', 'right']:
            save = getattr(self, attr)
            setattr(self, attr, getattr(other, attr))
            setattr(other, attr, save)

        # Fix what we broke
        self.up = other
        other.down = self

        # Introduce ourselves to our new neighbours
        self.update_neighbours()
        other.update_neighbours()
          
        # Swap in parent 1D arrays
        self.array_1d[self.position.x] = self
        other.array_1d[other.position.x] = other
        
    # Iterators through the 2D array
    
    def update_neighbours(self):
        if self.up:
            self.up.down = self
        if self.down:
            self.down.up = self
        if self.left:
            self.left.right = self
        if self.right:
            self.right.left = self
        
    def iter_up(self):
        while self:
            yield self
            self = self.up

    def iter_down(self):
        while self:
            yield self
            self = self.down

    def iter_left(self):
        while self:
            yield self
            self = self.left

    def iter_right(self):
        while self:
            yield self
            self = self.right

class Array1D(list):
    def __init__(self, array_2d, y, row):
        self.array_2d = array_2d
        self.y = y
        super().__init__(row)
        
        for x, member in enumerate(row):
            member.position = Position(x, y)
            member.array_1d = self
            
class Array2D(list):
    # 2D array that must be initialized all in one go (no appending/extending)
    def __init__(self, rows):
        super().__init__(Array1D(self, y, row) for y, row in enumerate(rows))
        
        # Set member left and right links
        for row in self:
            for left, right in zip(row, row[1:]):
                left.right = right
                right.left = left
                
        # Set member up and down links
        for row_above, row_below in zip(self, self[1:]):
            for up, down in zip(row_above, row_below):
                up.down = down
                down.up = up

        self.size = Size(len(max(self, key=len)), len(self))
        
    def flatten(self):
        # Convenience function to process all members with one iterator
        return itertools.chain.from_iterable(self)
    
    def member_at(self, position):
        # Convenience function to index with coordinate tuple
        return self[position[1]][position[0]]
    
if __name__ == '__main__':
    class T(Member):
        def __init__(self, value):
            self.value=value
            super().__init__()
        def __repr__(self):
            return self.value + super().__str__()
        
    array = Array2D([[T('a'), T('b')], [T('c'), T('d'), T('e')], [T('f'), T('g'), T('h')], [T('i'), T('j'), T('k')]])
    for row in array:
        print(row)
        
    array[1][1].swap_below()
    print()
    for row in array:
        print(row)
    print(array.size)
        
    array[1][1].swap_below()
    print()
    for row in array:
        print(row)
    print(array.size)

    array[0][0].swap_below()
    print()
    for row in array:
        print(row)
    print(array.size)
        
    array[0][0].swap_below()
    print()
    for row in array:
        print(row)
    print(array.size)

    array[1][1].unlink()
    print()
    for row in array:
        print(row)
    print(array.size)



from .array_2d import Member
from .coords import Size

class Element(Member):
    # Character that expects to be in an Array2D representing the original source
    def __init__(self, char):
        if not isinstance(char, str) or len(char) != 1:
            raise TypeError('Elements must be initialized with a str of length 1')
        self.char = char
        self.processed = False
        super().__init__()
            
    @classmethod
    def from_str(cls, string):
        # Helper function to make list of characters from a string
        return [cls(letter) for letter in string]
                
    def ispair(self):
        return not self.processed and self.right and not self.right.processed
    
    def row_processed(self, width, should_be):
        # Mark elements to the right processed, but only if they are should_be
        # print(f'Setting {width} elements to processed if they are {should_be}')
        for _ in range(width):
            if not self:
                raise ValueError('Went off end of row trying to mark it processed')
                break
            # print(f'  Checking to see if {self.char} matches')
            if self.char == should_be:
                # print(f'  Setting {self} to processed')
                self.processed = True
            self = self.right
    
    def __str__(self):
        if self.up:
            up = f'^{self.up.position.x},{self.up.position.y}'
        else:
            up = '^None'
        if self.down:
            down = f'v{self.down.position.x},{self.down.position.y}'
        else:
            down = 'vNone'
        if self.left:
            left = f'<{self.left.position.x},{self.left.position.y}'
        else:
            left = '<None'
        if self.right:
            right = f'>{self.right.position.x},{self.right.position.y}'
        else:
            right = '>None'
            
        return f'"{str(self.char)}"@{self.atstr()} {up} {down} {left} {right})'
    
    def get_part_size(self, top, middle, half_bottom):
        # Find the size of a part that has been extended with
        # special extender characters. Mark the extenders we use as processed
         
        if not self.ispair():
            raise AssertionError('Must not call on processed/missing elements')
        width = 2
        height = 2
        
        # Find length of top row of part
        this = self.right.right
        while this and not this.processed and this.char == top:
            width += 1
            this = this.right
        
        # Mark top row processed (but not the part pair itself)
        if width > 2:
            self.right.right.row_processed(width - 2, top)
        
        # Find height of part left size, marking rows processed as we go
        this = self.down
        while this and not this.processed and this.char == middle:
            height += 2
            next_down = this.down
            this.row_processed(width, middle)
            this = next_down
        
        # Find if we have an extra 1-high half-bottom, and mark it processed
        if this and not this.processed and this.char == half_bottom:
            height += 1
            this.row_processed(width, half_bottom)

        # print(Size(width, height))
        return Size(width, height)



#!/usr/bin/env python
from copy import copy, deepcopy
from colour import Color as Colour
 
class ColourList:
    # Either a text marker, or a list of colours, with "none" as an allowable colour.
    # Once colour lists are created, they are not expected to change.
    # Stringized version of colours are designed to be used in SVG documents, which
    # is why "none" is added to what colour supports natively.
    # Text markers (create by passing the literal "text") are a bit of a hack, to
    # allow fritizingpcb to use the same character to designate a cell to be either a
    # part with a colour, or a text string (the value of which is defined elsewhere
    # in the input YAML).
 
    def __init__(self, sequence):
        if not sequence:
            raise ValueError('ColourList must be initiated with a value')
 
        # print(type(sequence), sequence)
        self.text = False
        self.data = []
 
        if isinstance(sequence, str):
            sequence = [s.strip() for s in sequence.split(',')]
 
        if sequence[0] == 'text':
            self.text = True
            return

        for element in sequence:
            if element != 'none':
                self.data.append(Colour(element))
            else:
                self.data.append('none')
 
        if len(self.data) < 2:
            self.data.append(deepcopy(self.data[0]))
            if self.data[1] != 'none':
                try:
                    self.data[1].luminance -= 0.1
                    # Out-of-bounds luminance is not raised until colour is evaluated
                    # so force an evaluation to trigger ValueError
                    str(self.data[1])
                except ValueError:
                    self.data[1].luminance = 0
 
        if len(self.data) < 3:
            self.data.append(deepcopy(self.data[0]))
            if self.data[2] != 'none':
                try:
                    self.data[2].luminance += 0.1
                    str(self.data[2])
                except ValueError:
                    self.data[2].luminance = 1
 
    def __getitem__(self, item):
        if self.text:
            raise TypeError('Cannot index text marker')
        else:
            return self.data[item]
 
    def __str__(self):
        if self.text:
            raise TypeError('Cannot stringize text marker')
        return str(self.data[0])
 
    def __repr__(self):
        if self.text:
            return f'{self.__class__.__name__}("text")'
        else:
            datastr = ', '.join('"'+str(c)+'"' for c in self.data)
            return f'{self.__class__.__name__}([{datastr}])'
 
    def istext(self):
        return self.text
 
if __name__ == '__main__':
    x = 'none'
    print(f'\nTrying {x}')
    c = ColourList(x)
    if not c.istext():
        print(c, c[0], c[1], c[2])
    print(repr(c))
    x = 'text'
    print(f'\nTrying {x}')
    c = ColourList(x)
    if not c.istext():
        print(c, c[0], c[1], c[2])
    print(repr(c))
   
    x = 'blue'
    print(f'\nTrying {x}')
    c = ColourList(x)
    if not c.istext():
        print(c, c[0], c[1], c[2])
    print(repr(c))
   
    x = 'white'
    print(f'\nTrying {x}')
    c = ColourList(x)
    if not c.istext():
        print(c, c[0], c[1], c[2])
    print(repr(c))
 
    x = ['red', 'green', 'blue', 'yellow', 'cyan', 'magenta']
    print(f'\nTrying {x}')
    c = ColourList(x)
    if not c.istext():
        print(c, c[0], c[1], c[2])
    print(repr(c))
 
 
    x = 'red, green,blue ,yellow, cyan, none, magenta'
    print(f'\nTrying {x}')
    c = ColourList(x)
    if not c.istext():
        print(c, c[0], c[1], c[2])
    print(repr(c))



from .coords import Size
from .grid_drawing import GridDrawing

def pcb_svg(board, gridlines=False):
    # Draw PCN SVG, returning it in ElementTree XML
    grid = GridDrawing(
        size=board.size, 
        border=Size(0.2, 0.2),
        grid_size=board.config['metadata']['grid_size'],
        grid_unit=board.config['metadata']['grid_unit'],
        scale=board.config['metadata']['scale'],
        parts_dict=board.config['pcb-parts'])

    
#     c2 = CoordsTo(grid.border, Position.zero(), self.scale)
#     # print(Outline(board.layout).path_d(c2, Vector(0.1, 0.1), Size(0.3, 0.3), Size(0.6, 0.6)))
#     
# 
#     # c2 = CoordsTo(grid.border, Size.zero(), self.scale)
#     # path_d = outline.path_d(c2, Size.zero(), Size(0.3, 0.3), Size(0.6, 0.6))
#     # grid.add(grid.path(stroke='black', fill='none', stroke_width=lw/10*self.scale, d=path_d))
    silkscreen = grid.g(id='silkscreen')
    pcb_shape = grid.pcb_outline(board)
    silkscreen.add(pcb_shape)
    grid.add(silkscreen)

    copper1 = grid.g(id='copper1')
    copper0 = grid.g(id='copper0')

    font = board.config['appearance']['pcb']['font']
    colours = board.config['colours']
    
    for cell in board.connectors.values():
        copper0.add(grid.draw_part(cell, font, colours))

    # pcb.add(grid.grid_lines(*board.size))
    copper1.add(copper0)
    grid.add(copper1)

    if gridlines:
        grid.add(grid.grid_lines(*board.size))

    return grid.get_xml()



import math
from .coords import Dir, Vector, Size, CoordsTo, Position
from .helpers import find_part_code_from_key, find_colour_from_key

class PCBShapeMixin:
    # Mixin extending GridDrawing with the ability
    # to generate SVG to draw the PCB for the breadboard image
    #
    # Boards has the layout array, colours, and parts
    # We know border and scale
    
    def pcb_outline(self, board):
        stroke_width = board.config['appearance']['pcb']['outline']
        stroke_colour_key  =  board.config['appearance']['pcb']['outline-colour']
        stroke = board.colour(stroke_colour_key)
        
        outline_path = OutlinePath(board, self.parts_dict, holes=False)

        path_d = outline_path.to_d(self.border, Size.zero(), self.scale)
        outline = self.path(id='PCB-outline',
                            stroke=stroke, fill='none',
                            stroke_width=stroke_width * self.scale,
                            stroke_linejoin='round', d=path_d)

        return outline
    
    def pcb_shape(self, board):
        
        stroke_width = board.config['appearance']['breadboard']['highlight']
        offset = Size(stroke_width, stroke_width) / 2
        
        # board.print_array()
        outline_path = OutlinePath(board, self.parts_dict)
        # print(outline_path.dirs)

        # print(outline_path)
        pcb_colour = board.config['appearance']['breadboard']['PCB-colour']
        highlight = board.colour((pcb_colour, 2))
        lowlight = board.colour((pcb_colour, 1))
        fill = board.colour((pcb_colour, 0))
        
        group = self.g(id='PCB-board')
        
        path_d = outline_path.to_d(self.border, Size.zero(), self.scale)
        group.add(self.path(stroke=fill, fill='none',
                            stroke_width=stroke_width * 2 * self.scale,
                            stroke_linejoin='round', d=path_d))

        path_d = outline_path.to_d(self.border, -offset, self.scale)
        group.add(self.path(stroke=highlight, fill='none',
                            stroke_width=stroke_width * self.scale,
                            stroke_linejoin='round', d=path_d))

        path_d = outline_path.to_d(self.border, offset, self.scale)
        group.add(self.path(stroke=lowlight, fill='none',
                            stroke_width=stroke_width * self.scale,
                            stroke_linejoin='round', d=path_d))

        path_d = outline_path.to_d(self.border, Size.zero(), self.scale)
        group.add(self.path(stroke='none', fill=fill, d=path_d))
               
        return group
        
class OutlinePath:
    def __init__(self, board, parts_dict, holes=True):
        self.board = board
        self.dirs = []
        self.holes = []
        # self.layout = board.config['layout']
        # self.notpcb = board.config['specials']['not-pcb']
        
        # Find a top left corner
        # board.source is a 2D layout of elements, with
        # element.char == config['specials']['empty'] for non-pcb parts
        # print(board.layout.source)
        for element in board.source.flatten():
            # print(f"Is {element} != '{board.config['specials']['empty']}'")
            if element.char != board.config['specials']['empty']:
                self.start = element
                self.dir_right(element)
                break
            
        # FIXME look for actual holes (cells with ' '), not just hole parts

        if holes:
            # Find the hole part paths
            for cell in board.parts:
                part_key, part_ops = find_part_code_from_key(parts_dict,
                                                                cell.part_key)
                if part_key != None:
                    for shape in part_key['shapes']:
                        if shape['type'] == 'holepath':
                            self.holes.append((cell, part_ops, shape['args']))
                            
            for cell in board.connectors.values():
                part_key, part_ops = find_part_code_from_key(parts_dict,
                                                                cell.part_key)
                if part_key != None:
                    for shape in part_key['shapes']:
                        if shape['type'] == 'holepath':
                            self.holes.append((cell, part_ops, shape['args']))
            
        # for hole in self.holes:
        #     print(f'Hole: {hole}')
        
    def __str__(self):
        return f'start = {self.start}\ndirs = {self.dirs}'

    def to_d(self, border, offset, scale):
        c2 = CoordsTo(border, offset, scale)
        d = ['M', *c2.cabs(self.start.position * Vector(1,2))]

        # FIXME only need to do this once per outline, not once per to_d()
        merged = self.merge_dirs(self.dirs)
        
        for this, count in merged[:-1]:
            if this == Dir.RIGHT:
                d.extend(['l', *c2.crel(Vector.one().right() * count)])
            elif this == Dir.DOWN:
                d.extend(['l', *c2.crel(Vector.two().down() * count)])
            elif this == Dir.LEFT:
                d.extend(['l', *c2.crel(Vector.one().left() * count)])
            elif this == Dir.UP:
                d.extend(['l', *c2.crel(Vector.two().up() * count)])
            else:
                raise AssertionError("Can't happen")
        d.append('z')

        for cell, part_ops, args in self.holes:
            c2 = CoordsTo(border, cell.position+offset, scale, part_size=cell.size, ops=part_ops)
            d.extend(c2.expand_args_list(args))
            
        return d
    
    @staticmethod
    def merge_dirs(dirs):
        # Turn [RIGHT, RIGHT, DOWN, RIGHT, DOWN, DOWN, DOWN] into
        #      [(RIGHT, 2), (DOWN, 1), (RIGHT, 1), (DOWN, 3)]
        merged = [(dirs[0], 1)]
        for dir in dirs[1:]:
            if merged[-1][0] == dir:
                merged[-1] = (dir, merged[-1][1]+1)
            else:
                merged.append((dir, 1))
        return merged
    
    def dir_right(self, element):
        # Arrived a top left corner
        # This could be the start
        if element.up:
            self.dir_up(element.up)
        elif element.right:
            self.dirs.extend([Dir.RIGHT])
            self.dir_right(element.right)
        elif element.down:
            self.dirs.extend([Dir.RIGHT, Dir.DOWN])
            self.dir_down(element.down)
        elif element.left:
            self.dirs.extend([Dir.RIGHT, Dir.DOWN, Dir.LEFT])
            self.dir_left(element.left)
        else:
            self.dirs.extend([Dir.RIGHT, Dir.DOWN, Dir.LEFT, Dir.UP])
            
    def dir_down(self, element):
        # Arrived a top right corner
        # Can't go down to the start
        if element.right:
            self.dir_right(element.right)
        elif element.down:
            self.dirs.extend([Dir.DOWN])
            self.dir_down(element.down)
        elif element.left:
            self.dirs.extend([Dir.DOWN, Dir.LEFT])
            self.dir_left(element.left)
        elif element.up:
            self.dirs.extend([Dir.DOWN, Dir.LEFT, Dir.UP])
            self.dir_up(element.up)
        else:
            raise AssertionError("Can't happen")

    def dir_left(self, element):
        # Arrived a bottom right corner
        # Can arrive at the start going left, if PCB is one unit high
        if element.down:
            self.dir_down(element.down)
        elif element.left:
            self.dirs.extend([Dir.LEFT])
            self.dir_left(element.left)
        elif element.up:
            self.dirs.extend([Dir.LEFT, Dir.UP])
            self.dir_up(element.up)
        elif element.right:
            self.dirs.extend([Dir.LEFT, Dir.UP])
            if element == self.start:
                return
            self.dirs.extend([Dir.RIGHT])
            self.dir_right(element.right)
        else:
            raise AssertionError("Can't happen")

    def dir_up(self, element):
        # Arrived a bottom left corner
        if element.left:
            self.dir_left(element.left)
        elif element.up:
            self.dirs.extend([Dir.UP])
            self.dir_up(element.up)
        elif element.right:
            self.dirs.extend([Dir.UP]) # Finish left edge
            if element == self.start:
                return
            self.dirs.extend([Dir.RIGHT])
            self.dir_right(element.right)
        elif element.down:
            self.dirs.extend([Dir.UP])
            if element == self.start:
                return
            self.dirs.extend([Dir.RIGHT, Dir.DOWN])
            self.dir_down(element.down)
        else:
            raise AssertionError("Can't happen")



import xml.etree.ElementTree as ET

def fzp_filename(real_filename):
    (svg, dirname, filename) = real_filename.split('.', maxsplit=2)
    return f'{dirname}/{filename}'

def fzp_xml(board, input_filename, schematic_filename, pcb_filename, breadboard_filename):
    # Build Fritzing metadata, returning it in ElementTree XML
    
    schematic_filename = fzp_filename(schematic_filename)
    pcb_filename = fzp_filename(pcb_filename)
    breadboard_filename = fzp_filename(breadboard_filename)

    metadata =  board.config['metadata']
    fzp = ET.Element('module',
                        referenceFile=input_filename,
                        moduleId=metadata['SVGbasename'],
                        fritzingVersion='0.3.16b.02.24.4002')

    ET.SubElement(fzp, 'author').text = metadata['author']
    ET.SubElement(fzp, 'version').text = metadata['version']
    ET.SubElement(fzp, 'title').text = metadata['title']
    ET.SubElement(fzp, 'url').text = metadata['url']
    ET.SubElement(fzp, 'label').text = metadata['label']
    ET.SubElement(fzp, 'date').text = metadata['date']
    ET.SubElement(fzp, 'description').text = metadata['description']
    
    tag_list = ET.SubElement(fzp, 'tags')
    for tag in board.config['tags']:
        ET.SubElement(tag_list, 'tag').text = tag
        
    property_list = ET.SubElement(fzp, 'properties')
    for name, value in board.config['properties'].items():
        # print(key)
        ET.SubElement(property_list, 'property', name=name).text = value
        
    view_list = ET.SubElement(fzp, 'views')
    
    view_element = ET.SubElement(view_list, 'iconView')
    layer_list = ET.SubElement(view_element, "layers", image=breadboard_filename)
    ET.SubElement(layer_list, 'layer', layerId='icon')

    view_element = ET.SubElement(view_list, 'breadboardView')
    layer_list = ET.SubElement(view_element, "layers", image=breadboard_filename)
    ET.SubElement(layer_list, 'layer', layerId='breadboard')
    
    view_element = ET.SubElement(view_list, 'schematicView')
    layer_list = ET.SubElement(view_element, "layers", image=schematic_filename)
    ET.SubElement(layer_list, 'layer', layerId='schematic')
    
    view_element = ET.SubElement(view_list, 'pcbView')
    layer_list = ET.SubElement(view_element, "layers", image=pcb_filename)
    ET.SubElement(layer_list, 'layer', layerId='copper0')
    ET.SubElement(layer_list, 'layer', layerId='copper1')
    ET.SubElement(layer_list, 'layer', layerId='silkscreen')
    
    connector_list = ET.SubElement(fzp, "connectors")
    for cid, connector in enumerate(board.connectors.values()):
        connector_element = ET.SubElement(
            connector_list,
            'connector',
            type=connector.gender,
            name=f'pin {connector.number}',
            id=f'connector{cid}')
        
        if connector.description:
            text = f'{connector.label} {connector.description}'
        else:
            text = f'{connector.label}'
        ET.SubElement(connector_element, 'description').text = text
        
        view_list = ET.SubElement(connector_element, 'views')
        
        view_element = ET.SubElement(view_list, 'breadboardView')
        ET.SubElement(
            view_element, 'p',
            svgId=f'connector{cid}pin',
            layer='breadboard')
        
        view_element = ET.SubElement(view_list, 'schematicView')
        ET.SubElement(
            view_element, 'p',
            svgId=f'connector{cid}pin',
            layer='schematic',
            terminalId=f'connector{cid}terminal')
        
        view_element = ET.SubElement(view_list, 'pcbView')
        ET.SubElement(
            view_element, 'p',
            svgId=f'connector{cid}pin',
            layer='copper0')
        ET.SubElement(
            view_element, 'p',
            svgId=f'connector{cid}pin',
            layer='copper1')
        
    bus_list = ET.SubElement(fzp, "buses")
    for bus_name, cell_list in board.buses.items():
        bus_element = ET.SubElement(bus_list, 'bus', id=bus_name)
        for cell in cell_list:
            ET.SubElement(
                bus_element, 'nodeMember',
                connectorId=f'connector{cell.cid}')
        
    ET.SubElement(fzp, "schematic-subparts")
    ET.SubElement(fzp, "spice")
    
    return fzp



from .coords import Size
from .grid_drawing import GridDrawing

def breadboard_svg(board, gridlines=False):
    # Draw breadboard SVG, returning it in ElementTree XML
    grid = GridDrawing(
        size=board.size, 
        border=Size(0.2, 0.2),
        grid_size=board.config['metadata']['grid_size'],
        grid_unit=board.config['metadata']['grid_unit'],
        scale=board.config['metadata']['scale'],
        parts_dict=board.config['breadboard-parts'])

    breadboard = grid.g(id='breadboard')
    
#     c2 = CoordsTo(grid.border, Position.zero(), self.scale)
#     # print(Outline(board.layout).path_d(c2, Vector(0.1, 0.1), Size(0.3, 0.3), Size(0.6, 0.6)))
#     
# 
#     # c2 = CoordsTo(grid.border, Size.zero(), self.scale)
#     # path_d = outline.path_d(c2, Size.zero(), Size(0.3, 0.3), Size(0.6, 0.6))
#     # grid.add(grid.path(stroke='black', fill='none', stroke_width=lw/10*self.scale, d=path_d))
    pcb_shape = grid.pcb_shape(board)
    breadboard.add(pcb_shape)

    font = board.config['appearance']['breadboard']['font']
    colour_dict = board.config['colours']
    
    for cell in board.parts:
        breadboard.add(grid.draw_part(cell, font, colour_dict))

    # print(board.layout.connectors)
    for cell in board.connectors.values():
        breadboard.add(grid.draw_part(cell, font, colour_dict))

    for cell in board.text.values():
        breadboard.add(grid.draw_text(cell, board.text_dict, colour_dict))

    grid.add(breadboard)

    if gridlines:
        grid.add(grid.grid_lines(*board.size))
        
    return grid.get_xml()



from .array_2d import Member
from .coords import Size

class Cell(Member):
    def __init__(self, element):
        super().__init__()
        self.element = element
        
class EmptyCell(Cell):
    pass

class PCBCell(Cell):
    pass

class ConnectorCell(Cell):
    def __init__(self, element):
        # Caller only knows the connector number when created
        if not element.ispair():
            raise AssertionError('Must not call on processed/missing elements')
        self.number = int(element.char + element.right.char, base=10)
        super().__init__(element)
        self.colour_key =  None  # We don't know what kind of connector this is
        self.part_key = None # Board will parse that from connector descriptions
        self.size = Size(2, 2)
        self.schemed = None # Not yet positioned in schematic
        
    def update(self, colour_key, part_key, label, 
               description, gender, cid, part_ops=[]):
        # Caller uses update when all details of connector are known
        self.colour_key = colour_key # Gets
        self.part_key = part_key
        self.label = label
        self.description = description
        self.gender = gender
        self.cid = cid
        self.part_ops = part_ops
        
class PartCell(Cell):
    def __init__(self, element, size):
        if not element.ispair():
            raise AssertionError('Must not call on processed/missing elements')
        super().__init__(element)
        self.colour_key = element.char.upper()
        self.part_key = element.right.char
        self.size = size
        self.part_ops = []
                
class TextCell(Cell):
    def __init__(self, element):
        if not element.ispair():
            raise AssertionError('Must not call on processed/missing elements')
        super().__init__(element)
        self.text_marker = element.char
        self.text_key = element.char.upper() + element.right.char
        self.size = Size(2, 2)
# 
#     def __repr__(self):
#         if self.type == CellType.NONE:
#             return '     '
#         elif self.type == CellType.CONNECTOR:
#             return f'{self.number:02}{self.size.width}x{self.size.height}@{self.position.x},{self.position.y}'
#         elif self.type == CellType.PART:
#             return f'{self.colour_key}{self.part_key}{self.size.width}x{self.size.height}@{self.position.x},{self.position.y}'
#         elif self.type == CellType.TEXT:
#             return f'{self.text_key}{self.size.width}x{self.size.height}@{self.position.x},{self.position.y}'
#         elif self.type == CellType.PCB:
#             return f'PCB@{self.position.x},{self.position.y}'
#         elif self.type == CellType.NONE:
#             return f'None@{self.position.x},{self.position.y}'
#         else:
#             raise AssertionError("Can't happen")
#         
# 



import svgwrite

from .coords import Size, CoordsTo, Position, Vector
from .pcb_shape_mixin import PCBShapeMixin
from .helpers import find_part_code_from_key, find_colour_from_key
from .cell import ConnectorCell, PartCell

import logging
log = logging.getLogger(__name__)

class GridDrawing(svgwrite.Drawing, PCBShapeMixin):
    # Construct an SVG drawing based on a grid
    # All input uses a cell size of 1.0x1.0
    # All output elements use a cell size of scale (e.g 7.2 x 7.2)
    # The grid can have a border, in which case cells can be drawn outsize the grid
    # The SVG w/h will be written in svgwrite unit (if specified)
    units = {
        'mm': svgwrite.mm, 
        'cm': svgwrite.cm, 
        'inch': svgwrite.inch,
        'px': svgwrite.px,
        'none': 1}

    def __init__(self, size, grid_size, grid_unit, scale, parts_dict, border=Size(0, 0)):
        
        # self.filename = filename
        self.size = size
        self.grid_size = grid_size
        self.grid_unit = grid_unit
        self.scale = scale
        self.parts_dict = parts_dict
        self.border = border
        self.coords_to = CoordsTo(border, Position(0, 0), scale)
        
        try:
            SVG_size = ((size + border * 2) * grid_size) * self.units[grid_unit]
        except KeyError:
            raise ValueError('Unrecognized unit: {grid_units}')
        
        viewbox_size = (size + border * 2) * self.scale
        viewbox_string = f'0 0 {int(viewbox_size.width)} {int(viewbox_size.height)}'
        
        super().__init__(
            profile='tiny',
            debug=True,
            size=SVG_size,
            viewBox=viewbox_string
        )
        
    def __repr__(self):
        return f'GridDrawing(size={self.size}, grid_size={self.grid_size}, grid_units={self.grid_units}, scale={self.scale}, border={self.border})'
    
    def grid_lines(self, width, height):
        c2 = CoordsTo(self.border, 0, self.scale)
        group = self.g(id='grid-lines')
        
        for x in range(self.size.width + 1):
            stroke_dasharray="none"
            if x % 2 == 1:
                stroke_dasharray="18,6"
            group.add(self.line(
                c2.cabs(Position(x, 0)), 
                c2.cabs(Position(x, self.size.height)),
                stroke_dasharray=stroke_dasharray,
                stroke='magenta', stroke_width=1))
        for y in range(self.size.height + 1):
            stroke_dasharray="none"
            if y % 2 == 1:
                stroke_dasharray="8,16"
            group.add(self.line(
                c2.cabs(Position(0, y)),
                c2.cabs(Position(self.size.width, y)),
                stroke_dasharray=stroke_dasharray,
                stroke='magenta', stroke_width=1))
        
        return group
    
    def draw_text(self, cell, text_dict, colour_dict):
        c2 = CoordsTo(self.border, cell.position, self.scale, cell.size)
        text = text_dict[cell.text_key]
        # print(f'text', text)
        # print(f'text["text"]: "{text["text"]}"')
        x, y = c2.pabs(Position(0.5, 0.5)) # Rotate about center of part
        fill = find_colour_from_key(colour_dict, text['colour'], None)
        text_svg = self.text(
                text['text'],
                # insert=c2.prel(Position(0.5, 0.5)),
                insert=c2.prel(Size(-0.5, text['size'] * 0.18)),
                fill=fill,
                font_family=text['font'],
                font_size=c2.clen(text['size']),
                text_anchor='start',
                transform=f'translate({x},{y}) rotate({text["angle"]})')
                # transform=f'rotate({text["angle"]})')
        text_svg.update({'xml:space': 'preserve'})
        return text_svg
    
    def draw_holepath(self, *args, **kwargs):
        pass
    
    def draw_ellipse(self, group, id_dict, fill, stroke, stroke_width, expargs):
        # print('Ellipse', fill, stroke, stroke_width, expargs)
        if len(expargs) != 4:
            raise ValueError(f'Ellipses must have 4 args: {expargs}')
        group.add(self.ellipse(
            **id_dict,
            center=(expargs[0], expargs[1]),
            r=(abs(expargs[2]), abs(expargs[3])),
            fill=fill,
            stroke=stroke,
            stroke_width=stroke_width))

    def draw_circle(self, group, id_dict, fill, stroke, stroke_width, expargs):
        # print('Circle', fill, stroke, stroke_width, expargs)
        if len(expargs) != 3:
            raise ValueError(f'Circles must have 3 args: {expargs}')
        group.add(self.circle(
            **id_dict,
            center=(expargs[0], expargs[1]),
            r=(abs(expargs[2])),
            fill=fill,
            stroke=stroke,
            stroke_width=stroke_width))

    def draw_rect(self, group, id_dict, fill, stroke, stroke_width, expargs):
        # print('Rect', fill, stroke, stroke_width, expargs)
        if len(expargs) != 4:
            raise ValueError(f'Rectangles must have 4 args: {expargs}')
        # if str(fill) == '#cc1414':
        #     print(fill, stroke, stroke_width, expargs)
        
        minx = min(expargs[0], expargs[0]+expargs[2])
        miny = min(expargs[1], expargs[1]+expargs[3])
        group.add(self.rect(
            **id_dict,
            insert=(minx, miny),
            size=(abs(expargs[2]), abs(expargs[3])),
            fill=fill,
            stroke=stroke,
            stroke_width=stroke_width))

    def draw_rectround(self, group, id_dict, fill, stroke, stroke_width, expargs):
        # print('Rect', fill, stroke, stroke_width, expargs)
        if len(expargs) != 6:
            raise ValueError(f'Rounded rectangles must have 6 args: {expargs}')
        minx = min(expargs[0], expargs[0]+expargs[2])
        miny = min(expargs[1], expargs[1]+expargs[3])
        group.add(self.rect(
            **id_dict,
            insert=(minx, miny),
            size=(abs(expargs[2]), abs(expargs[3])),
            rx=abs(expargs[4]), ry=abs(expargs[5]),
            fill=fill,
            stroke=stroke,
            stroke_width=stroke_width))

    def draw_path(self, group, id_dict, fill, stroke, stroke_width, expargs):
        # print('Path', fill, stroke, stroke_width, expargs)
        group.add(self.path(
            **id_dict,
            d=expargs,
            fill=fill,
            stroke=stroke,
            stroke_width=stroke_width))

    drawers = {
        'holepath': draw_holepath,
        'ellipse': draw_ellipse,
        'circle': draw_circle,
        'rect': draw_rect,
        'rectround': draw_rectround,
        'path': draw_path,
    }
    
    ops_dict = {
        'draw-in': [],
        'flip-to': ['flip'],
        'flop-to': ['flop'],
        'flipflop-to': ['flip', 'flop'],
        'transpose-to': ['transpose'],
        'transpose-flip-to': ['transpose', 'flip'],
        'transpose-flop-to': ['transpose', 'flop'],
        'transpose-flipflop-to': ['transpose', 'flip', 'flop']
    }
    
    def shadow_draw(self, drawer, c2, shadow, group, id_dict, fill,
                    stroke, stroke_width, args):
        # log.info(f'Drawing shape with shadow {shadow}, fill {fill}, type {type(fill)}')
        if str(fill) == 'none' and str(stroke) == 'none':
            return
        
        if shadow:
            if stroke != 'none':
                log.info(f'Shadowing shapes with stroke "{stroke}" does not work well')
            
            try:
                offset = Vector(shadow[0], shadow[1])
            except TypeError:
                offset = Vector(shadow, shadow)

            # print('before transpose_only', offset)
            offset = c2.transpose_only(offset)
            # print('after transpose_only', offset)

            if offset[0] == offset[1]:
                # For equal shadows, using the stroke to draw looks nicer at the
                # corners, and produces less SVG
                shadow_stroke_width = c2.clen(offset[0])
                
                expargs = c2.expand_args_list(args)
                drawer(self, group, id_dict, 'none', fill[0],
                       shadow_stroke_width*2, expargs)

                expargs = c2.offset(offset/2).expand_args_list(args)
                drawer(self, group, id_dict, 'none', fill[1],      
                       shadow_stroke_width, expargs)

                expargs = c2.offset(-offset/2).expand_args_list(args)
                drawer(self, group, id_dict, 'none', fill[2],
                       shadow_stroke_width, expargs)
            else:
                # For unequal shadows, we have to use fill to draw shadow, as
                # SVG strokes can't have different widths and heights (like
                # a wide calligraphy pen)
                if offset[0] == 0:
                    shadow_vectors = [Vector(0, 1)]
                    corner_vectors = []
                elif offset[1] == 0:
                    shadow_vectors = [Vector(1, 0)]
                    corner_vectors= []
                else:
                    shadow_vectors = [Vector(1,1), Vector(0,1), Vector(1,0)]
                    corner_vectors = [Vector(1,-1), Vector(-1,1)]
                    
                # print('offset', offset)
                # print('shadow_vectors', shadow_vectors)
                # print('corner_vectors', corner_vectors)
                for v in corner_vectors:
                    expargs = c2.offset(offset*v).expand_args_list(args)
                    drawer(self, group, id_dict, fill[0], 'none',
                           0, expargs)
                
                for v in shadow_vectors:
                    expargs = c2.offset(offset*v).expand_args_list(args)
                    drawer(self, group, id_dict, fill[1], 'none',
                           0, expargs)
                
                for v in shadow_vectors:
                    expargs = c2.offset(-offset*v).expand_args_list(args)
                    drawer(self, group, id_dict, fill[2], 'none',
                           0, expargs)

        expargs = c2.expand_args_list(args)
        drawer(self, group, id_dict, fill, stroke, stroke_width, expargs)
        
    def draw_part(self, cell, font, colour_dict):
        # Draw part or connector
        
        # Both have the part_key set
        part_code, part_ops = find_part_code_from_key(self.parts_dict, cell.part_key)
        # colour = find_colour_from_key(colour_dict, cell.colour_key, None)
        
        part_ops.extend(cell.part_ops)
            
        if isinstance(cell, ConnectorCell):
            gid = f'pin{cell.number}'
            name = f'{cell.number:<2}'
            # print(part_ops)
        else:
            try:
                name = part_code['name']
            except KeyError:
                # Part does not have name
                name = cell.colour_key + cell.part_key
            except TypeError:
                # Part is None (not known)
                name = cell.colour_key + cell.part_key
            x, y = cell.element.position
            w, h = cell.size
            gid = f'{name}-cell:{x}.{y}-size:{w}x{h}'
            
        group = self.g(id=gid)
        
        if part_code:
  
            for shape in part_code['shapes']:
                c2 = CoordsTo(self.border, cell.position, self.scale, Size.one())
                fill, stroke, stroke_width =  self.style(shape, colour_dict,
                                                         cell.colour_key, c2)

                try:
                    drawer = self.drawers[shape['type']]
                except KeyError:
                    raise ValueError(f'Unknown shape type: {shape["type"]}')

                try:
                    id_dict = {'id': f'connector{cell.cid}{shape["id"]}'}
                except (AttributeError, KeyError):
                    id_dict = {}

                # print(shape)
                try:
                    shadow = shape['shadow']
                except KeyError:
                    shadow = 0

                if 'draw-in' in shape:
                    try:
                        step = int(shape['step'])
                    except KeyError:
                        step = 1
                        
                    part_cells = self.part_cells(cell, part_ops, step)
                    # print('Part cells by type')
                    # for k, v in part_cells.items():
                    #     l = '|'.join([str(c) for c in v])
                        # print(f'{k} {l}')
                        
                    for op_name, ops_list in self.ops_dict.items():
                        if op_name in shape:
                            for location in shape[op_name]:
                                for op_cell in part_cells[location]:
                                    c2 = CoordsTo(self.border, op_cell.position,
                                                  self.scale, cell.size, 
                                                  ops=ops_list+part_ops, step=step)
                                    self.shadow_draw(drawer, c2, shadow, group,
                                                id_dict, fill, stroke,
                                                stroke_width, shape['args'])

                else:
                    c2 = CoordsTo(self.border, cell.position, self.scale,
                                  cell.size, ops=part_ops)
                    # expargs =  c2.expand_args_list(shape['args'])
                    # if cell.string == 'c-':
                    #     print('XXXX', repr(cell))
                    #     print('XXXX', shape['args'])
                    #     print('XXXX', expargs)
                    #     print('XXXX', c2)
                    # drawer(self, group, id_dict, fill, stroke, stroke_width, expargs)
                    self.shadow_draw(drawer, c2, shadow, group,
                                     id_dict, fill, stroke,
                                     stroke_width, shape['args'])
            
        else:
            fill = find_colour_from_key(colour_dict, '?', cell.colour_key)
            c2 = CoordsTo(self.border, cell.position, self.scale, cell.size)
            position = c2.cabs(Position(0.1, 0.1))
            size = c2.jump(Size(0.8, 0.8))
            stroke_width = c2.clen(0.05)
            font_size = 1
            group.add(self.rect(
                position,
                size,
                fill=fill,
                stroke='magenta',
                stroke_width=stroke_width))
            group.add(self.text(
                name,
                c2.pabs(Position(0.5, 0.5)) + c2.crel(Size(0, font_size * 0.36)),
                fill='magenta',
                font_family = font,
                font_size=c2.clen(font_size),
                text_anchor='middle'
                ))
        return group
    
    @staticmethod
    def part_array(cell):
        # Make up an sub-array of the cell's Array2D that make up this part
        # FIXME Will probably barf for an offset part on bottom row
        array = []
        row_start = cell
        for _ in range(cell.size.height):
            row = []
            part_cell = row_start
            for _ in range(cell.size.width):
                row.append(part_cell)
                part_cell = part_cell.right
            array.append(row)
            row_start = row_start.down
        return array

    @classmethod
    def part_cells(cls, cell, part_ops, step=1):
        if not isinstance(cell, PartCell):
            raise AssertionError(f'Cell is not a part: {cell}')

        result = {}
        array = cls.part_array(cell)
        
        if 'transpose' in part_ops:
            array = [list(x) for x in zip(*array)]
            
        result['top-left'] = [array[0][0]]
        result['top-middle'] = array[0][step:-step:step]
        result['top-right'] = [array[0][-step]]
        
        result['middle-left'] = []
        result['middle'] = []
        result['middle-right'] = []

        for row in array[step:-step:step]:
            result['middle-left'].append(row[0])            
            result['middle'].extend(row[step:-step:step])
            result['middle-right'].append(row[-step])
            
        result['bottom-left'] = [array[-step][0]]
        result['bottom-middle'] = array[-step][step:-step:step]
        result['bottom-right'] = [array[-step][-step]]
        
        return result
    
    @staticmethod
    def style(shape, colour_dict, cell_colour_key, c2):
            
        try:
            fill_colour_key = shape['fill']
        except KeyError:
            fill = 'none'
        else:
            fill = find_colour_from_key(colour_dict, fill_colour_key,
                                            cell_colour_key)
        try:
            stroke_colour = shape['stroke']
        except KeyError:
            stroke = 'none'
        else:
            stroke = find_colour_from_key(colour_dict, stroke_colour,
                                              cell_colour_key)

        try:
            stroke_width = shape['stroke-width']
        except KeyError:
            stroke_width = 0
        else:
            stroke_width = c2.clen(stroke_width)
        
        return fill, stroke, stroke_width
        



#!/usr/bin/env python

from enum import Enum, auto
from copy import deepcopy

from .array_2d import Array2D
from .cell import Cell, ConnectorCell, TextCell, PartCell, PCBCell, EmptyCell
from .element import Element
from .coords import Vector, Size

class Layout(Array2D):
    # 2D array of cells, which can each be none, a part or a connector
    # Built from list of strings showing PCB layout
    # Layout[y][x] references the cell at a given position
    # Other attibutes are:
    #   source:     2D array of source characters (used to draw PCB outline)
    #   size:       the layout size
    #   connectors: dict of connectors. indexed by number
    #   parts:      list of parts
    
    def __init__(self, source, empty_char, pcb_chars, text_chars,
                 top, middle, half_bottom):
        # We remove from this, as cells are indentified
        
        # for row in source:
        #     for element in row:
        #         print(f'Element {element}')        # Remove stuff not part of the PCB
        # for element in source.flatten():
        #     if element.char == empty_char:
        #         element.remove()

        # This copy we keep, as a record of the PCB shape
        # self.source = deepcopy(source)
        
        # # Layout is twice as high as source, and last row could be offset down
        # initializer = []
        # for _ in range(source.size.height * 2):
        #     initializer.append([Cell() for _ in range(source.size.width)])
        # super().__init__(initializer)
        
        # Remove stuff in the PCB that isn't a connector, part, or text
        # for element in source.flatten():
        #     if element.char and element.char in pcb_chars:
        #         cell = self.obj_by_pos(element.layout_position())
        #         cell.make_pcb(element)
        #         element.remove()
            
        cell_rows = []
        for row in source:
            cell_upper_row = []
            cell_lower_row = []
            for element in row:
                # print(f'Checking element {element}')
                if element.char == empty_char:
                    cell_upper_row.append(EmptyCell(element))
                    cell_lower_row.append(EmptyCell(element))
                    element.processed = True
                elif element.processed or element.char in pcb_chars:
                    cell_upper_row.append(PCBCell(element))
                    cell_lower_row.append(PCBCell(element))
                    element.processed = True
                elif element.right and not element.right.char in pcb_chars:
                    # Find connectors
                    try:
                        cell = ConnectorCell(element)
                        cell_upper_row.append(cell)
                        cell_lower_row.append(PCBCell(element))
                    except ValueError:
                        # Must be text or part
                        # Part extenders from higher parts will have been processed
                        if element.char.upper() in text_chars:
                            cell = TextCell(element)
                        else:
                            # get_part_size() marks extenders as processed
                            size = element.get_part_size(top, middle, half_bottom)
                            cell = PartCell(element, size)
                            
                        if element.char.isupper():
                            cell_upper_row.append(cell)
                            cell_lower_row.append(PCBCell(element))
                        else:
                            cell_upper_row.append(PCBCell(element))
                            cell_lower_row.append(cell)
                            
                    element.processed = True
                    element.right.processed = True                        
                else:
                    raise ValueError(f'Unexpected character {element.char} at {element.position}')
            cell_rows.append(cell_upper_row)
            cell_rows.append(cell_lower_row)
        
        # print(cell_rows)
        
        super().__init__(cell_rows)

    def __str__(self):
        rows = []
        for row in self:
            rows.append('|'.join(str(c) for c in row))
        return '\n'.join(rows)
        
        
if __name__ == '__main__':
    source = [
        '.Aa+.', 
        'Ca+..', 
        '###..', 
        '.bB..',
        '.##ff']
    layout = Layout(source, ' ', '.', '+', '#', '=')
    print(layout)
   
